﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using System.Reflection;

namespace System
{
  public static class TypeExtension
  {
    public static PropertyInfo GetProperty(this Type type, string propertyName, bool inheritance)
    {
      PropertyInfo value = type.GetProperty(propertyName);
      if (inheritance && value == null && type.IsInterface)
      {
        foreach (Type interfaceType in type.GetInterfaces())
          return GetProperty(interfaceType, propertyName, inheritance);
      }
      return value;
    }

    public static IEnumerable<InheritedInterfaceMappingElt> GetInterfaceMap(this Type type, Type interfaceType, bool interfaceInheritance)
    {
        foreach (var inheritedInterfaceMappingElt in GetInterfaceMap(type, interfaceType))
            yield return inheritedInterfaceMappingElt;
        if (interfaceInheritance)
            foreach (Type interfaceSubType in interfaceType.GetInterfaces())
                foreach (var inheritedInterfaceMappingElt in GetInterfaceMap(type, interfaceSubType))
                    yield return inheritedInterfaceMappingElt;
    }
    private static IEnumerable<InheritedInterfaceMappingElt> GetInterfaceMap(this Type type, Type interfaceType)
    {
        var map = type.GetInterfaceMap(interfaceType);
        int nbMethods = map.InterfaceMethods.Length;
        for (int indexMethod = 0; indexMethod < nbMethods; indexMethod++)
        {
            yield return new InheritedInterfaceMappingElt
            {
                InterfaceMethod = map.InterfaceMethods[indexMethod],
                InterfaceType = map.InterfaceType,
                TargetMethod = map.TargetMethods[indexMethod],
                TargetType = map.TargetType
            };
        }
    }

    public static bool IsNullable(this Type t)
    {
        return !t.IsSubclassOf(typeof(ValueType)) || t.IsGenericType && t.GetGenericTypeDefinition() == typeof(Nullable<>);
    }

  }

  public class InheritedInterfaceMappingElt
  {
      private MethodInfo _interfaceMethod;
      private Type _interfaceType;
      private MethodInfo _targetMethod;
      private Type _targetType;

      public MethodInfo InterfaceMethod
      {
          get { return _interfaceMethod; }
          internal set { _interfaceMethod = value; }
      }
      public PropertyInfo InterfaceProperty
      {
          get { return PropertyInfoHelper.GetPropertyInfoFromGetOrSetMethod(InterfaceType, InterfaceMethod.Name); }
      }
      public Type InterfaceType
      {
          get { return _interfaceType; }
          internal set { _interfaceType = value; }
      }
      public MethodInfo TargetMethod
      {
          get { return _targetMethod; }
          internal set { _targetMethod = value; }
      }
      public PropertyInfo TargetProperty
      {
          get { return PropertyInfoHelper.GetPropertyInfoFromGetOrSetMethod(TargetType, TargetMethod.Name); }
      }
      public Type TargetType
      {
          get { return _targetType; }
          internal set { _targetType = value; }
      }
  }

  public static class PropertyInfoHelper
  {
      private static string GetPropertyNameFromGetOrSetMethod(string getOrSetMethodName, bool onlySet)
      {
          if (getOrSetMethodName == null)
              return null;
          var lastDot = getOrSetMethodName.LastIndexOf('.');
          if (lastDot == getOrSetMethodName.Length - 1)
              return null;
          var value = getOrSetMethodName.Substring(lastDot + 1);
          if (!((!onlySet && value.StartsWith("get_")) || value.StartsWith("set_")))
              return null;
          value = value.Remove(0, 4);
          return getOrSetMethodName.Substring(0, lastDot + 1) + value;
      }
      public static PropertyInfo GetPropertyInfoFromGetOrSetMethod<T>(string getOrSetMethodName)
      {
          return GetPropertyInfoFromGetOrSetMethod(typeof(T), getOrSetMethodName);
      }
      public static PropertyInfo GetPropertyInfoFromGetOrSetMethod(Type type, string getOrSetMethodName)
      {
          var propertyName = GetPropertyNameFromGetOrSetMethod(getOrSetMethodName, false);
          if (propertyName == null)
              return null;
          return type.GetProperty(propertyName);
      }
      public static PropertyInfo GetPropertyInfoFromSetMethod<T>(string setMethodName)
      {
          return GetPropertyInfoFromSetMethod(typeof(T), setMethodName);
      }
      public static PropertyInfo GetPropertyInfoFromSetMethod(Type type, string setMethodName)
      {
          var propertyName = GetPropertyNameFromGetOrSetMethod(setMethodName, true);
          if (propertyName == null)
              return null;
          return type.GetProperty(propertyName);
      }
  }

  /// <see cref="http://msdn2.microsoft.com/fr-fr/library/bb727332.aspx"/>
  public static class ReflectionHelper<T>
  {
      /// <exception cref="System.ArgumentNullException"></exception>
      /// <exception cref="System.ArgumentException"></exception>
      public static PropertyInfo GetProperty(Expression<Func<T, object>> selector)
      {
          if (selector == null)
              throw new ArgumentNullException();
          Expression e = selector.Body;
          while (!((e is MemberExpression) &&
            (e.NodeType == ExpressionType.MemberAccess) &&
            ((e as MemberExpression).Member is PropertyInfo)))
          {
              if (e is UnaryExpression)
              {
                  switch (e.NodeType)
                  {
                      case ExpressionType.Convert:
                      case ExpressionType.ConvertChecked:
                      case ExpressionType.TypeAs:
                          e = (e as UnaryExpression).Operand;
                          break;
                      default:
                          throw new ArgumentException();
                  }
              }
              else
                  throw new ArgumentException();
          }
          return (e as MemberExpression).Member as PropertyInfo;
      }
  }
}
